/* 
 * generate.c --- written by yasuko matsubara 2012.7.4
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "nrutil.h"
#include "tool.h"
#include <unistd.h>
#define BUFSIZE 256
#define ZERO 0.00000001

void _normalize();
void _estimate();
void _generate();

int main(int argc,char **argv){
   int     i,j,t;
   int     wantSamp = 0;            
   int     wantEst  = 0;            
   int     objectID =-1;
   int     argcheck = 0;
   int	   ch; 
   extern char	*optarg;        
   extern int	optind, opterr; 
   while ((ch = getopt(argc, argv, "sae:")) != -1)
     switch (ch){
       case 'e': // count estimation (only for one object)
         wantEst=1;  argcheck++; 
         objectID=   atoi(optarg)-1;
         break;
       case 's': // event-set generation
         wantSamp=1; argcheck++;
         break;
       case 'a': // count estimation (for all objects)
         wantEst=1;  argcheck++;
         break;
     } //switch
   argc -= optind;
   argv += optind;

   if(argcheck!=1){ 
     fprintf(stderr, "choose one option -i or -s or -a \n");
     fprintf(stderr, "   -i <objectID> (count estimation) \n");
     fprintf(stderr, "   -a            (count estimation for all objects) \n");
     fprintf(stderr, "   -s            (event-set generation) \n");
   }//if

   if(argc != 9 || argcheck!=1) 
      error("Usage: Ofn Afn Cfn x_ifn k u v n outdir[-i objectID/-s/-a]", "generate");

   char *Ofn = argv[0];
   char *Afn = argv[1];
   char *Cfn = argv[2];
   char *x_ifn = argv[3];
   int  k = atoi(argv[4]);
   int  u = atoi(argv[5]);
   int  v = atoi(argv[6]);
   int  n = atoi(argv[7]);
   char *outdir = argv[8];
   char fn[BUFSIZE];
   fprintf(stderr, "u=%d, v=%d n=%d k=%d\n", u, v, n, k);

   fprintf(stderr, "Alloc matrix ...\n");
   // alloc matrix O A C
   double **O =  (double**) dmatrix(0, k, 0, u);
   double **A =  (double**) dmatrix(0, k, 0, v);
   double **C =  (double**) dmatrix(0, k, 0, n);
   double *cumP = (double*) dvector(0, v);
  
   // alloc x_i[u], x_ijt[u][v][n] 
   float   *x_i   = (float*)vector(0, u);
   // if x_ijt[u][v][n] > memory size, please use -e option 
   int uu = u;
   if(objectID != -1)
     uu = 1;
   float ***x_ijt = (float***)malloc(sizeof(float**)*uu);
   for(i=0;i<uu;i++)
     x_ijt[i] = (float**)matrix(0,v,0,n);

   fprintf(stderr, "load data ...\n");
   // read O & A & C matrix
   readMatrix(Ofn, O, k, u);
   readMatrix(Afn, A, k, v);
   readMatrix(Cfn, C, k, n);
   readfVector(x_ifn, x_i, u);

   fprintf(stderr, "Normalize params ...\n");
   _normalize(O, A, C, k, u, v, n);
 
   /// default setting
   if(wantEst && objectID != -1){
     fprintf(stderr, "Count estimtion ...(objectID=%d)\n", objectID+1);
         _estimate(objectID, O, A, C, x_i, x_ijt[0], k, u, v, n);
     sprintf(fn, "%sgen.url%d", outdir, objectID+1);
     writefMatrix(fn, x_ijt[0], v, n);
   }

   /// default setting for all objects
   if(wantEst && objectID == -1){
     fprintf(stderr, "Count estimtion for all objects ...\n");
       for(i=0;i<uu;i++)
         _estimate(i, O, A, C, x_i, x_ijt[i], k, u, v, n);
     fprintf(stderr, "write x_ijt, x_ijt ...\n");
     write_f_xijt(outdir, x_ijt, uu, v, n);
   }


   /// if you want to generate realistic events
   if(wantSamp){
     fprintf(stderr, "event-set generation... \n");
     _generate(O, A, C, cumP, x_i, x_ijt, k, u, v, n);
     fprintf(stderr, "write x_ijt, to twd format ...\n");
     write_f_twd(outdir, x_ijt, u, v, n);
   }

}


void _estimate(
int objectID, 
double **O, 
double **A, 
double **C, 
float  *x_i, 
float **x_jt,
int k, int u, int v, int n){

  int j,t,r;
  double val;
  for(j=0;j<v;j++)
    for(t=0;t<n;t++){
      val = 0.0;
      for(r=0;r<k;r++)
        val+=O[r][objectID]*A[r][j]*C[r][t];
      val *= (double)n * (double)x_i[objectID];
      x_jt[j][t] = (float) val;
    }//t

}

void _normalize(
double **O, double **A, double **C,
int k, int u, int v, int n){
  int i,r;
  double sum;
  // O
  for(i=0;i<u;i++){
    sum = ZERO;
    for(r=0;r<k;r++)
      sum += O[r][i];
    for(r=0;r<k;r++)
      O[r][i]/=sum;
  }//i
  // A
  for(r=0;r<k;r++){
    sum=ZERO;
    for(i=0;i<v;i++)
      sum += A[r][i];
    for(i=0;i<v;i++)
      A[r][i]/=sum; 
  }//r
  // C
  for(r=0;r<k;r++){
    sum=ZERO;
    for(i=0;i<n;i++)
      sum += C[r][i];
    for(i=0;i<n;i++)
      C[r][i]/=sum; 
  }//r
}

int _draw(double *cumP, int k){
     int i;
     for(i=1;i<k;i++)
       cumP[i]  += cumP[i-1];
     double max  = cumP[k-1];     
     double rand = max * random()/ RAND_MAX ;
     for(i=0;i<k;i++)
       if(cumP[i] >= rand)
         return i;
     error("cannot assign", "_draw()");
     return -1;
}

void _generate(
double **O, 
double **A, 
double **C, 
double *cumP, 
float  *x_i, 
float  ***x_ijt,
int k, int u, int v, int n){
   int i,j,t,r, x;

   for(i=0;i<u;i++)
     for(j=0;j<v;j++)
       for(t=0;t<n;t++)
         x_ijt[i][j][t] = 0;

   fprintf(stderr, "generate object (i=1, ..., %d) \ni=", u);
   for(i=0;i<u;i++){

     fprintf(stderr, "%d ", i+1);
     for(x=0;x<x_i[i]*n;x++){
       /// draw a hidden z (O)
       for(r=0;r<k;r++)
         cumP[r] = O[r][i];
       int z = _draw(cumP, k);
       /// draw an actor (A)
       for(j=0;j<v;j++)
         cumP[j] = A[z][j];
       int e = _draw(cumP, v);
       /// draw a timestamp (C)
       for(t=0;t<n;t++)
         cumP[t] = C[z][t];
       int t = _draw(cumP, n);
       /// add {i, e, t}
       x_ijt[i][e][t]++;
     }//x

   }//i
   fprintf(stderr, "\n");
}

